import numpy as np
import pandas as pd
import traceback

from adapter.comm.excel import Book, xl2pd, pd2xl
from adapter.comm.sql import Sql

from adapter.comm.tools import process_column_labels

import re, sys

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


from pdb import set_trace as bp


class Excel(object):
    """Loads bulk named tables
    and ranges from .xlsx to python as
    a dictionary of dataframes.

    Parameters:

        file_path: str
            Path to an excel sheet

        pre_existing_keys: dictionary key index
            Keys is the previously loaded
            dictionary of dataframes
    """

    def __init__(self, file_path, pre_existing_keys=None):

        self.wb = Book(file_path)
        self.file_path = file_path
        self.pre_existing_keys = pre_existing_keys

        log.info("Connected to: {}".format(file_path))

    def load(
        self,
        data_object_names=None,
        kind="all",
    ):
        """Opens the file provided
        through file_path, loads
        all or a subset of tables
        and named ranges
        available in the file and
        converts to a python format.

        Parameters:

            data_object_names: list
                List of string data object names
                to load. Data objects can be
                named tables and named ranges, see
                'type' kwarg for details.
                Default: None means
                that all data objects found in the
                input file get loaded.

            kind: str
                'tables' : gets only named tables
                'ranges' : gets only named ranges
                'all' : gest both named tables
                and named ranges

        Returns:

            dict_of_dfs: dict of dataframes
                Python dictionary with table name
                keys and input table values.
        """
        dict_of_dfs = self.get_named_data_objects(data_object_names, kind=kind)

        return dict_of_dfs

    def get_named_data_objects(self, data_object_names, kind="all"):
        """Grabs data defined as excel tables, named ranges, or both.

        Parameters:

            data_object_names: list or None
                List of string data object names
                to load. Data objects can be
                named tables and named ranges, see
                'type' kwarg for details.
                Default: None means
                that all data objects found in the
                input file get loaded.

            kind: str
                'tables' : gets only named tables
                'ranges' : gets only named ranges
                'all' : gest both named tables
                and named ranges

        Returns:

            dict_of_dfs: dict of dataframes
                Python dictionary with table name/
                named range as keys and contents of
                the corresponding named data object
                values.
        """
        if kind == "tables":
            # grab all the input tables/named ranges
            all_input_objects = self.wb.tables

        if kind == "ranges":
            all_input_objects = {
                x.name: x
                for x in self.wb.names
                if "_FilterDatabase" not in x.name and "_xlfn." not in x.name
            }

        if kind == "all":
            all_input_tables = self.wb.tables
            all_input_ranges = {
                x.name: x
                for x in self.wb.names
                if "_FilterDatabase" not in x.name and "_xlfn." not in x.name
            }

            all_input_tables.update(all_input_ranges)

            all_input_objects = all_input_tables.copy()

        elif kind not in ["all", "ranges", "tables"]:
            msg = "An unsupported value provided for kwarg kind {}."
            log.error(msg.format(kind))
            raise ValueError

        if self.pre_existing_keys is not None:

            Debugger.check_for_duplicates(
                self.pre_existing_keys, all_input_objects.keys()
            )

        # initiate dictionaries of input table dataframes
        # and populate
        dict_of_dfs = dict()

        if len(all_input_objects.keys()) > 0:

            if type(data_object_names) == list:
                for data_object_name in data_object_names:
                    if data_object_name not in all_input_objects.keys():
                        msg = "{} not found in the input file {}."
                        log.error(msg.format(data_object_name, self.file_path))
                        raise ValueError

            elif (
                (data_object_names is None)
                or (data_object_names is np.nan)
                or (np.isnan(data_object_names))
            ):
                data_object_names = all_input_objects.keys()

            else:
                msg = "Unsupported type ({}) passed for table names, {}."
                log.error(msg.format(type(table_names), table_names))
                raise ValueError

            try:

                for data_object_name in data_object_names:

                    try:
                        # prepare labels (strip extra spaces)
                        dict_of_dfs[
                            data_object_name
                        ] = self.wb.named_range_to_df(
                            all_input_objects[data_object_name],
                            verbose=True,
                        )

                        dict_of_dfs[
                            data_object_name
                        ].columns = process_column_labels(
                            dict_of_dfs[data_object_name].columns
                        )

                    except:
                        msg = (
                            "Failed to read input table named {}"
                            " from input file {}. "
                            "If the data contained in the table "
                            "is needed in the analysis please attempt"
                            " to rename the table/range using strings and numerals "
                            "and/or further check the data range or table "
                            "definition."
                        )

                        log.warning(
                            msg.format(data_object_name, self.file_path)
                        )
                        raise ValueError

                msg = "Read in input tables and ranges from {}."
                log.info(msg.format(self.file_path))

            except:
                msg = "Failed to read input tables from {}."

                log.error(msg.format(self.file_path))
                raise ValueError

        else:
            msg = (
                "Neither named tables nor named ranges found in "
                "the input file {}."
            )
            log.info(msg.format(self.file_path))

        return dict_of_dfs


class Db(object):
    """Loads tables from a database to python
    as a dictionary of dataframes.

    Parameters:

        file_path: str
            Path to an db file

        pre_existing_keys: dictionary key index
            Keys is the previously loaded
            dictionary of dataframes
    """

    def __init__(self, file_path, pre_existing_keys=None):

        self.db = Sql(file_path)
        self.file_path = file_path
        self.pre_existing_keys = pre_existing_keys

    def load(self, table_names=None):
        """Loads tables

        Parameters:

            table_names: list
                List of table names to load.
                Default: None = load all tables
        """
        try:
            all_dict_of_dfs = self.db.tables2dict(close=True)

            dict_of_dfs = dict()

            if len(all_dict_of_dfs.keys()) > 0:

                if type(table_names) == list:

                    for table_name in table_names:

                        if table_name not in all_dict_of_dfs.keys():
                            msg = "{} not found in the input file {}."
                            log.error(msg.format(table_name, self.file_path))
                            raise ValueError

                        else:
                            dict_of_dfs[table_name] = all_dict_of_dfs[
                                table_name
                            ]
                else:
                    dict_of_dfs = all_dict_of_dfs

            if self.pre_existing_keys is not None:

                Debugger.check_for_duplicates(
                    self.pre_existing_keys, dict_of_dfs.keys()
                )

        except:
            msg = "Failed to read input tables from {}."
            log.error(msg.format(self.inpath))
            raise ValueError

        return dict_of_dfs


class Db_sqlalchemy(object):
    """Loads tables from a database using sqlalchemy to python
    as a dictionary of dataframes.

    Parameters:

        file_path: str
            Path to an db file

        pre_existing_keys: dictionary key index
            Keys is the previously loaded
            dictionary of dataframes
    """

    def __init__(self, file_path, pre_existing_keys=None):
        # Imports in this class so if you don;t need them it will still work
        try:
            from adapter.Secret import database_credentials
        except:
            raise ValueError(
                """You must define your "Secret.py" file within the adapter folder.  
                                Replace the example "Secret_example.py" with your credentials"""
            )
        from sqlalchemy import create_engine

        self.cxn_str = (
            "postgresql://"
            + database_credentials["Username"]
            + ":"
            + database_credentials["Password"]
            + "@"
            + file_path
        )
        self.engine = create_engine(self.cxn_str)
        self.file_path = file_path
        self.pre_existing_keys = pre_existing_keys

    def load(self, table_names=None):
        """Loads tables

        Parameters:

            table_names: list
                List of table names to load.
                Default: None = load all tables
        """

        try:
            dict_of_dfs = dict()
            for table_name in table_names:
                sql_string = "SELECT * FROM {table}".format(table=table_name)
                dict_of_dfs[table_name] = pd.read_sql(
                    con=self.engine, sql=sql_string
                )

            if self.pre_existing_keys is not None:

                Debugger.check_for_duplicates(
                    self.pre_existing_keys, dict_of_dfs.keys()
                )

        except:
            traceback.print_exc()
            msg = "Failed to read input tables from {}."
            log.error(msg.format(self.file_path))
            raise ValueError

        return dict_of_dfs


class Debugger(object):
    @staticmethod
    def check_for_duplicates(pre_existing_keys, table_names):
        """Checks if among data tables already
        added to the dict_of_dfs there exists
        a table of a same name with any of the
        tables being added to the dict_of_dfs.

        Parameters:

            pre_existing_keys: dictionary key index
                Keys is the previously loaded
                dictionary of dataframes

            table_names: list or string
                List of table names; or a
                single string table name.
        """
        msg = (
            "An identically name table/range"
            " {} was already read in. "
            "Please rename all tables and ranges"
            " in the inputs uniquely. "
            "It may be that the same-named "
            "table came from a different input file."
        )

        if not isinstance(table_names, str):
            for name in table_names:
                if name in pre_existing_keys:
                    log.error(msg.format(name))
                    raise ValueError(msg.format(name))

        else:
            try:
                if pre_existing_keys is not None:
                    if table_names in pre_existing_keys:
                        log.error(msg.format(name))
                        raise ValueError
            except:
                msg = "Unexpected table_name ({}) format."
                log.error(msg.format(table_names))



class Labels(object):
    """Label maps related to
    functional input tables parsed
    by the adapter.
    """

    def __init__(self):
        pass

    def set_labels(self):
        """Table names and column labels"""
        self.labels = {
            # table names
            "run_pars": "run_parameters",
            "extra_files": "inputs_from_files",
            # column labels
            "outpath": "Output Path",
            "version": "Version",
            "inpath": "File Path",
            "tbl_nam": "Table Name",
            "query": "Query Only",
        }

        return self.labels



# -*- coding: utf-8 -*-
"""
Initial version created and documented by Youness Bennani.
"""

import xlwings as xw
import pandas as pd
import re
from glob import glob
from functools import reduce

import sys


class Name(xw.main.Name):
    """Class used in the definition of the class ```Names```."""

    def expand(self):
        create_named_range(self.wb, self, resize=True)

    @property
    def wb(self):
        full_address = str(self.refers_to_range)
        wb_name = re.search(r"\[(.*)\]", full_address).groups()[0]
        wb = Book(wb_name)
        return wb


class Names(xw.main.Names):
    """Class used in the definition of the class ```Book```."""

    def __call__(self, name_or_index):
        return Name(impl=self.impl(name_or_index))

    def add(self, name, **kwargs):
        return Name(create_named_range(self.wb, name, **kwargs).name)


class Table(object):
    """Class used in the definition of the class ```Book```."""

    def __init__(self, name, refers_to, wb=None):
        self.name = name
        if not wb:
            wb = xw.apps.active.books.active
        self.wb = wb
        sheet_name, range_address = _parse_address(refers_to)
        if not sheet_name:
            sheet_name = wb.sheets.active.name
        self.refers_to_range = wb.sheets(sheet_name).range(range_address)
        self.refers_to = self.refers_to_range.__str__().split("]")[1][:-1]

    def __repr__(self):
        return "<Table '%s': %s>" % (self.name, self.refers_to)


class BookDataFrames(dict):
    """Object used to store Excel named ranges converted to dataframes."""

    pass


class Book(xw.main.Book):
    """Improved version of xlwings' ```Book``` object. This class allows for the
    support of Excel data tables, and circumvents other limitations."""

    def __init__(self, fullname=None, impl=None):
        super(Book, self).__init__(fullname, impl)
        self.dfs = BookDataFrames()

    @property
    def tables(self):
        return get_tables(self)

    @property
    def names(self):
        names = Names(impl=self.impl.names)
        names.wb = self
        return names

    def create_name(self, range_name, **kwargs):
        return create_named_range(self, range_name, **kwargs)

    def add_sheet(self, sheet_name):
        """Add sheet or return it if it already exists."""
        if sheet_name.lower() in (s.name.lower() for s in self.sheets):
            print("Sheet {} exists. Assigning it".format(sheet_name))
            return self.sheets(sheet_name)
        else:
            print("Added sheet {}.".format(sheet_name))
            return self.sheets.add(sheet_name)

    def range(self, address, sheet_name=None, verbose=True):
        return get_range(self, address, sheet_name, verbose)

    def all_names_to_df(
        self, keep_sheet_name=False, verbose=True, header_row=1, index_col=0
    ):
        """Turn all names in an Excel workbook to pandas dataframes."""
        # Excel includes solver-type objects as named ranges. Those shouldn't be
        # read in.
        named_ranges = [
            name for name in self.names if "solver" not in name.name
        ] + list(self.tables.values())
        could_not_read = []
        for named_range in named_ranges:
            try:
                self.named_range_to_df(
                    named_range,
                    keep_sheet_name=keep_sheet_name,
                    header_row=header_row,
                    index_col=index_col,
                    verbose=verbose,
                )
            except:
                print("Warning: couldn't read name {}".format(named_range))
                could_not_read += [named_range]
        return set(named_ranges) - set(could_not_read)

    def named_range_to_df(
        self,
        named_range,
        keep_sheet_name=False,
        sheet_name=None,
        header_row=0,
        index_col=0,
        verbose=True,
    ):
        """Turns a named range in Excel into a pandas dataframe.
        Also adds the dataframe to the ```dfs``` attribute.

        Args:
            named_range (xlwings.main.Name or Table):
                The named range to transfer.
            keep_sheet_name (bool): Whether to keep the name of the sheet in the
                name of the variable created.
            sheet_name (str): The name of the worksheet from which we want to
                get the range. This is ignored if the address contains a sheet
                name.
            index_col (int or None): The column number that should be used to
                create an index for the Pandas dataframe If the first column is
                the index, then set index_col=1 (not zero) The default value is
                index_col=None: all columns are read in as data, and a separate
                index is created for the dataframe.
            header_row (int or None): The row number that should be used to
                create column headings. Has similar functionality to index_col.
            verbose (bool): Prints warning messages if certain issues are
                encountered.

        Returns:
            pd.DataFrame: The Excel data turned into a Pandas DataFrame.
        """
        if isinstance(type(named_range), str):
            named_range = Name(get_named_range(self, named_range, sheet_name))

        # Determine what the name should be.
        if keep_sheet_name:
            # Keep the sheet name in the name of the dataframe.
            name = named_range.name.replace("!", "_")
        else:
            name = (
                named_range.name.split("!")[1]
                if named_range.name.find("!") > 0
                else named_range.name
            )

        rg = named_range.refers_to_range
        # Determine whether there is a header and an index.
        df_content = rg.value
        # If the named_range is 2D, assume it has a header and no index.
        if type(df_content) == list and type(df_content[0]) == list:

            # Get dataframe using ```xl2pd```.
            df = xl2pd(self, rg, header_row=1, index_col=index_col)

        # If named range is 1D
        # assume the first value is a header if it is a string, and no index
        if type(df_content) == list:
            if type(df_content[0]) == str:
                df = pd.DataFrame(df_content[1:], columns=[df_content[0]])
            # Otherwise just treat 1D array as data
            elif type(df_content[0]) != list and type(
                df_content[0] in [float, int, bool]
            ):
                df = pd.DataFrame(df_content)

        # If named range is a single value, use name of the range as header
        if type(df_content) != list and type(df_content) in [
            str,
            float,
            int,
            bool,
        ]:
            df = pd.DataFrame({rg.name.name: [df_content]})

        # Assign the dataframe to the name.
        if verbose:
            print("Read in {}".format(named_range))  # mg changed assigning to
            # read in
        exec("self.dfs.{} = df".format(name))
        self.dfs[name] = df

        return df


def _parse_address(full_address):
    """Parse a string containing a range address.

    Parse an Excel range address into a sheet name and a range specification.

    Example:
        ```_parse_address("'my sheet'!A1:X100")``` would return
        ```("my_sheet", "A1:X100")```.
    """
    splitaddr = full_address.split("!")
    if len(splitaddr) == 1:
        sheet_name = None
        address = splitaddr[0]
    elif len(splitaddr) == 2:
        characters_to_strip = list("'<>=")
        sheet_name = reduce(
            lambda x, y: x.strip(y), characters_to_strip, splitaddr[0]
        )
        sheet_name = sheet_name.split("]")[-1]
        address = splitaddr[1].strip(">")
    else:
        raise ValueError(
            "Invalid address in _parse_address: "
            + 'cannot contain multiple "!" characters'
        )
    return (sheet_name, address)


def get_tables(wb):
    """Get all the table names in ```wb```.

    Args:
        wb : xlwings workbook object.

    Returns:
        dict: A dictionary of strings where the keys are all the table names in
        ```wb```, and the values are the corresponding sheet names.
    """

    if sys.platform.lower() == "darwin":
        # the `api` attribute acts differently for windows and osx. (Linux undeveloped).
        # On top of that, OSX excel tables that implicitly includes headers require an [#all] tag when grabbing the range for the Table. Otherwise the first row of data gets read as headers
        # This apparently ONLY works for excel tables, and NOT named ranges on OSX
        return_dict = dict()
        for ws in wb.sheets:
            objects = ws.api.list_objects()
            if (
                objects
            ):  # Can be a mysterious "k.missingvalue" when there are no excel tables, which gets interpreted as False by bool
                for obj in objects:
                    return_dict[obj.name()] = Table(
                        obj.name(),  # name of table
                        ws.name
                        + "!"
                        + ws.range(
                            obj.name() + "[#all]"
                            if not obj.name().endswith("[#all]")
                            else obj.name()
                        ).address,
                        wb,
                    )  # Windows: obj.Name, OSX: obj.name(), obj.Range.Address -> ws.range(obj.name()).address

        return return_dict
    elif sys.platform.lower().startswith("win"):
        return {
            obj.Name: Table(obj.Name, ws.name + "!" + obj.Range.Address, wb)
            for ws in wb.sheets
            for obj in ws.api.ListObjects
            if ws.api.ListObjects.Count > 0
        }


def get_range(wb, address, sheet_name=None, verbose=True):
    """Get a range identified by ```address``` from the workbook ```wb```.

    Returns an xlwings range object, given a workbook, an address,
    and optionally a worksheet.
    If the worksheet is not specified, then the address must
    contain the worksheet name explicitly.

    Args:
        wb (xw.main.Book): An xlwings Book object.

        address (str): A string containing an Excel address
            (either like 'A1:B5' or explicitly containing a worksheet
            name like "'My worksheet'!C5:X73").
        sheet_name (str): The name of the worksheet from which we want to get
            the range. This is ignored if the address contains a sheet name.
        verbose (bool): Prints warning messages if certain issues are
            encountered.

    Returns:
        xw.main.Range: The range that has been found.
    """

    # Determine if the address contains sheet information
    match = re.search(r"^=*(.*)!(.*)", address)
    if match:
        sheet_name, address = match.groups()
        sheet_name = sheet_name.strip("'")
    elif not sheet_name:
        if verbose:
            sheet_name = wb.sheets[0].name
            print(
                "Warning: no explicit sheet information"
                " was passed to get_range. Assuming the desired"
                ' sheet is "{}".'.format(sheet_name)
            )

    # Otherwise we use the sheet_name keyword to get the sheet.
    try:
        rng = wb.sheets(sheet_name).range(address)
    except:
        if verbose:
            print(
                'Warning: unable to find a range at "'
                + address
                + '" in sheet "'
                + sheet_name
                + '".'
            )
        rng = None

    return rng


def get_named_range(wb, range_name, sheet_name=None, verbose=True, **kwargs):
    """Get a named range and return it as an xlwings range object.

    Args:
        wb: an xlwings workbook object
        range_name: a string containing the range name to be accessed.
        sheet_name: optional specification of the worksheet where the named
            range lives (for locally defined names, for example).
            If sheet is not specified, the code searches globally for the
            range name.
        verbose: print warning messages. (default=True)

    Returns:
        xw.main.Range: An xlwings range corresponding to the range name
            provided.
    """
    if not sheet_name:
        sheet_name = kwargs.get(sheet_name)
    try:
        if sys.platform.lower().startswith("win"):
            if sheet_name:
                obj = wb.sheets(sheet_name)
            else:
                obj = wb
            return obj.names[range_name].refers_to_range
        elif sys.platform.lower() == "darwin":
            # This is a ridiculous way to do it, I know. But if you try to get a range from a workbook by name, or sheet that doesn't have it, it breaks
            for ws in wb.sheets:
                for obj in ws.api.list_objects():
                    if obj.name() == range_name:
                        return ws.range(
                            range_name + "[#all]"
                            if not range_name.endswith("[#all]")
                            else range_name
                        )
    except:
        if verbose:
            print("Warning: unable to find a range named " + range_name + ".")
        return None


def create_named_range(
    wb,
    range_name,
    address=None,
    sheet_name=None,
    local=False,
    resize=False,
    verbose=True,
):
    """Create a named range.

    Create a named range OR re-create an existing named range, extending
    or contracting it as appropriate to fit the area that is filled with data.
    Resize or make name local.

    Args:
        wb (xw.main.Book): An xlwings workbook object
        range_name (str): The range name to create (code will search for an
            existing range by this name)
        address (str): The address over which to create the named range.
            Can be omitted if the named range already exists.
        sheet_name (str): The name of the worksheet on which to create (or
            search for) the named range. ignored if an address is passed with
            a sheet name included.
        local (bool): If True, the name will be defined as a local name to the
            sheet.
        resize (bool): if True, resize the range to cover the full region
            containing data (uses VBA CurrentRegion property), and apply the
             name to that range. (Default: False.)

    Returns:
        xw.main.Range: An xlwings range corresponding to the range name
            created."""
    rg = None

    if type(range_name) in (str, unicode):
        rg = get_named_range(wb, range_name)
    elif type(range_name) == xw.main.Range:
        rg, range_name = range_name, None
    elif type(range_name) in (xw.main.Name, Name):
        rg = range_name.refers_to_range
        range_name = rg.name.name

    if address:
        if rg:
            print(
                "A named range was found, but an address was provided."
                " Deleting existing named range."
            )
            rg.name.delete()
        sheet_name_from_address, address = _parse_address(address)
        # Use the user-provided sheet if available.
        rg = get_range(wb, address, sheet_name or sheet_name_from_address)

    if not rg:
        print(
            "Please provide a valid value for ```range_name```. If you're "
            "providing a string, if the name does not already exist, please "
            "also provide a valid address."
        )
        return

    # This is needed if the named range is to be local.
    if not sheet_name:
        sheet_name = rg.sheet.name

    if resize:
        rg = rg.current_region

    # Create the named range.
    rg.name = (
        sheet_name + "!" if local and "!" not in range_name else ""
    ) + range_name

    if verbose:
        info_string = " as a named range local to sheet," if local else ""
        print(
            "Created named range {},{} in sheet {}, at address {}.".format(
                range_name, info_string, sheet_name, rg.address
            )
        )

    return rg


def xl2pd(
    workbook, myrange, index_col=0, header_row=0, formulas=False, **kwargs
):
    """Turn a named range from an Excel workbook into a Pandas dataframe.

    Args:
        workbook (str, Book or xw.main.Book): Either the filename of a workbook
            to be read, or an xlwings Book object. For example, this:

            ```
            xl2pd('filename.xlsx', 'A1:C3', 'SheetName')
            ```

            is equivalent to this:

            ```
            wb = xlwings.Book('filename.xlsx')
            xl2pd(wb, 'A1:C3', 'SheetName')
            ```

        myrange (str or xw.main.Range): A string containing either a range in
            the usual Excel format (e.g., 'A1:C3'), or the name of a named range
            in the workbook being read (The code searches for a
            colon character, ':', to determine which kind of range
            this is In the unlikely event that you have a named range
            whose name contains a colon, then set named_range=True
            to override this.)

        index_col (int ): The column number that should be used to create
            an index for the Pandas dataframe If the first column is the index,
            then set index_col=1 (not zero) The default value is
            index_col=0: all columns are read in as data, and a separate
            index is created for the dataframe.

        header_row (int): The row number that should be used to create
            column headings. Has similar functionality to index_col.

        formulas (bool): If True, read in Excel formulas as strings. Note that
            this will read in the *entire* data set (including index and
            column entries) as unicode, rather than numerical data.
            (default behavior is to read in the values generated by
            the formulas and coerce to appropriate data types.)

        **kwargs (dict): Keyword arguments passed through to get_range or
            get_named_range

    Returns:
        pd.DataFrame: A Pandas dataframe with the data retrieved.
    """
    if isinstance(workbook, str):
        # Load workbook from specified file
        workbook = xw.Book(workbook)
    # (otherwise assume a Book object was passed.)

    if isinstance(myrange, str):
        if ":" in myrange:
            rng = get_range(workbook, myrange, **kwargs)
        else:
            rng = get_named_range(workbook, myrange, **kwargs)
    else:
        rng = myrange

    if rng:
        if formulas:
            data = rng.formula  # This is a tuple of tuples.
            cols = list(data[0])
            cols[0] = "convert_to_index"
            ret_df = pd.DataFrame(list(data[1:]), columns=cols)
            ret_df.set_index("convert_to_index", drop=True, inplace=True)
        else:
            if sys.platform.startswith("win"):
                # Convert directly to dataframe using xlwings
                ret_df = rng.options(
                    pd.DataFrame, header=header_row, index=index_col
                ).value

            elif sys.platform == "darwin":
                ### >>>>> The following code (with a few trivial renamings) comes directly from xlwings source code,
                # and is subject to copyright <<<<<
                # Copyright (c) 2020, Zoomer Analytics LLC
                # Subject to BSD-3 License for above copyright holder: https://opensource.org/licenses/BSD-3-Clause
                if header_row == 1:
                    columns = pd.Index(rng.value[0])
                elif header_row > 1:
                    columns = pd.MultiIndex.from_arrays(rng.value[:header])
                else:
                    columns = None

                ret_df = pd.DataFrame(
                    rng.value[header_row:],
                    columns=columns,
                    dtype=None,
                    copy=False,
                )

                # handle index by resetting the index to the index first columns
                # and renaming the index according to the name in the last row
                if index_col > 0:
                    # rename uniquely the index columns to some never used name for column
                    # we do not use the column name directly as it would cause issues if several
                    # columns have the same name
                    ret_df.columns = pd.Index(range(len(ret_df.columns)))
                    ret_df.set_index(
                        list(ret_df.columns)[:index_col], inplace=True
                    )

                    ret_df.index.names = pd.Index(
                        value[header_row - 1][:index_col]
                        if header_row
                        else [None] * index_col
                    )

                    if header_row:
                        ret_df.columns = columns[index_row:]
                    else:
                        ret_df.columns = pd.Index(range(len(df.columns)))

                ### >>>>> End of snippet that comes directly from xlwings <<<<<
    else:
        ret_df = None

    return ret_df


def pd2xl(
    dataframe,
    workbook,
    path=None,
    sheet_name="Sheet1",
    start_cell="A1",
    create_sheet=True,
    header=False,
    index=False,
    range_name=None,
    save=False,
    close=False,
    clear=False,
    resize=False,
    **kwargs
):
    """Write a Pandas dataframe to an Excel workbook.

    Args:

        dataframe (pd.DataFrame): A pandas DataFrame to be written into Excel

        workbook (str, Book or xw.main.Book): Either the filename of a workbook
            to be read, or an xlwings Book object. For example, this:

            ```
            xl2pd('filename.xlsx', 'A1:C3', 'SheetName')
            ```

            is equivalent to this:

            ```
            wb = xlwings.Book('filename.xlsx')
            xl2pd(wb, 'A1:C3', 'SheetName')
            ```

        sheet_name (str, default = 'Sheet1'): The name of the worksheet to which
            the dataframe will be written. This is ignored if a named range is
            used (so the named range is assumed to be globally scoped).

        create_sheet (bool, default:False): If specified sheet does not exist,
            create it.

        start_cell (str): Upper leftmost cell where DataFrame will be written.
            If header and index are both True, start_cell will be empty with
            index titles starting one cell below, and header names starting one
            cell to the right. Default is cell A1. This parameter is
            ignored if range_name is given.

        header (bool, default = False): If True will print column headers (i.e.
            dataframe.columns).

        index (bool, default = False): If True will print row names
            (i.e. dataframe.index).

        range_name (str): Named excel range for dataframe to be written. If this
            keyword is passed, then the code first tries to find
            and overwrite a named range by this name, ignoring the
            start_cell value. If no such name exists, the code writes
            out to the start_cell location and creates a new named
            range covering the output region.

        clear (bool): If True, clear the range to be written before writing.
            The range cleared is either the current named range, or
            the full contiguous region containing values to the left
            of and below the start_cell.

        resize (bool, default= False): If True, resize the named range to match
            the data being written.

        save (bool): If True, save the workbook before returning.

        path (str): Path to the output file where this workbook should be saved.
            If save=False, this is ignored. If this is not set, then
            the save location will be the same as for the workbook that
            is passed in.

        close (bool): If True, close the workbook before returning.

        **kwargs (dict): Keyword arguments passed through to get_range or
            get_named_range

    Returns:
        xw.main.Book: The workbook where the dataframe was saved.
    """

    if isinstance(type(workbook), str):
        # If the workbook doesn't exist, create it.
        if not glob(workbook):
            wb = xw.Book()
            wb.save(workbook)
            workbook = wb
        else:
            # Load workbook from specified file
            workbook = xw.Book(workbook)
    # (otherwise assume a Book object was passed.)

    rg = None
    if range_name:
        rg = get_named_range(workbook, range_name, **kwargs)

    if not rg:
        if create_sheet:
            # Create the sheet if it's not there.
            try:
                workbook.sheets.add(sheet_name)
            except ValueError:
                pass
        rg = get_range(workbook, start_cell, sheet_name, **kwargs)

    if clear:
        rg.current_region.clear()

    rg.options(index=index, header=header).value = dataframe

    if range_name:
        create_named_range(
            workbook,
            range_name,
            address=rg.address,
            sheet_name=rg.sheet.name,
            resize=resize,
        )

    if save:
        workbook.save(path)
    if close:
        workbook.close()

    return workbook



import os
import logging
import sqlite3

import pandas as pd

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

from pdb import set_trace as bp


class Sql(object):
    """Performs python-sqlite db communication.

    Parameters:

        path_OR_dbconn: str or a database connection instance
            Full path to a database file or an already
            instantiated connection object
    """

    def __init__(self, path_OR_dbconn):
        # recognize or create the connection object
        # *mig enable this for other db flavors, such
        # as sql db and postgres

        if type(path_OR_dbconn) == str:
            try:
                # for aboslute or mounted paths
                self.db = sqlite3.connect(path_OR_dbconn)
            except:
                # for paths relative to run path
                path = os.path.join(os.getcwd(), path_OR_dbconn)
                self.db = sqlite3.connect(path)

        elif type(path_OR_dbconn) == sqlite3.Connection:
            self.db = path_OR_dbconn
        else:
            log.error(
                "Neither a path to a db file, "
                "nor a database connection got passed."
            )
            raise ValueError

    def tables2dict(self, close=True):
        """Reads all tables contained in a
        sql database and converts them to a
        pandas dataframe.

        Parameters:

            close: boolean, default=True
                If True, closes the connection      to db

        Returns:

            data: dict of pandas dataframes
                Saves each of the sql tabels as a
                pandas dataframe under a sql table
                name as a key
        """

        cursor = self.db.cursor()
        cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()
        data = dict()
        for table_name in tables:
            table_name = table_name[0]
            data[table_name] = pd.read_sql_query(
                """ SELECT * FROM '{}' """.format(table_name), self.db
            )

        if close:
            self.db.close()

        return data

    def table2pd(self, table_name, column_label_row=0):
        """Reads in a single sql table.

        Parameters:

            table_name: str
                sql table name

            column_label_row: int, default=0
                Index of the row which gets
                converted into column labels

        Returns:

            df: pandas dataframe
                Sql table read in as a pandas df.
        """
        df = pd.read_sql(
            """ SELECT * FROM '{}' """.format(table_name),
            self.db,
            index_col=None,
        )

        return df

    def pd2table(self, df, table_name, close=False):
        """Write a dataframe out to the database.
        If same named table exists, it gets replaced

        Parameters:

            table_name: str
                sql table name

            close: boolean, default=False
                If True, closes the connection to db
        """

        df.to_sql(table_name, self.db, if_exists="replace", index=False)

        if close:
            self.db.close()

        return True

    def csv2table(
        self,
        path_to_csv,
        table_name,
        column_label_row=0,
        converters=None,
        close=False,
    ):
        """Use to update bulk price or performance data.
        If same named table exists, it gets replaced

        Parameters:

            path_to_csv: str
                Full path to the csv table

            table_name: str
                sql table name of choice

            column_label_row: int, default=0
                Index of the row which gets
                converted into column labels

            converters: dict, default=None
                According to pandas documentation:
                Dict of functions for converting
                values in columns. Keys can be integers
                or column labels.

            close: boolean, default=False
                If True, closes the connection to db
        """

        csv = pd.read_csv(
            path_to_csv, converters=converters, header=column_label_row
        )

        csv.to_sql(table_name, self.db, if_exists="replace", index=False)

        if close:
            self.db.close()

        return True

    def commit(self, sql_command, close=False):
        """Execute a custom sql command

        Parameters:

            sql_command: string
                sql_command to exectute

        Returns:

            close: boolean, default=False
                If True, closes the connection to db
        """

        self.db.cursor().execute(sql_command)
        self.db.commit()
        if close:
            self.db.close()

        return True



import os
import numpy as np
import pandas as pd

from adapter.to_python import Excel, Db, Db_sqlalchemy, Debugger
from adapter.label_map import Labels

from adapter.comm.tools import convert_network_drive_path

from datetime import datetime
import re
import sqlite3
from shutil import copy
import ntpath

import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

from pdb import set_trace as bp


class IO(object):
    """Connects to the main input
    file that can be an excel sheet,
    a csv file or a database,
    loads the data, looks for
    additional input data paths
    and loops through those to
    get data as well. Saves a
    full input DB and provides the
    user with output and db paths,
    and database connections. It allows
    for large tables to be only queried and
    not loaded in python.

    Parameters:

        path : string
            Path to the initial input table. An
            initial input table can be of
            any type (.xlsx, .csv, .db) and can
            contain pointers to further input
            files. It is recommended to have
            a `run_parameters` table (see test
            folders on the master branch
            of the adapter repo for examples)
            that provides an output path and a
            version substring.

        mapping: list
            list of 2-tuples where 0th entry of each 
            tuple is the name of a windows network drive 
            location (e.g. "A:") and the 1st entry is OSX 
            network drive location (e.g. "/Volumes/A"). 
            Defaults to [("X:","/Volumes/my_folder")].

    """

    def __init__(
        self, 
        path, 
        os_mapping=[("X:", "/Volumes/my_drive")]):

        path = convert_network_drive_path(path,
            mapping = os_mapping)

        self.input_path = path

        if isinstance(path, str):
            self.input_type = self.get_file_type(path)

        # set labels
        self.la = Labels().set_labels()

    def get_file_type(self, path):
        """Extracts the file type from the fullpath.

        Parameters:

            path : string
                File path
        """
        extns = re.split("\.", path)[-1]

        # extns = ...
        file_type = ""

        if extns == "xlsx":
            file_type += "excel"

        elif (extns == "db") or (extns == "sqlite"):
            file_type += "database"

            # *mig add more file extension checks
            # as needed

        elif extns == "csv":
            file_type += "text"

        # If the path contains lbl.gov, it is likely a database that can be used with sqlalchemy
        elif "lbl.gov" in path:
            file_type += "sqlalchemy"

        else:
            msg = "Passed an unsupported input file type: {}."
            log.error(msg.format(extns))

        return file_type

    def load(
        self,
        create_db=True,
        db_flavor="sqlite",
        close_db=True,
        save_input=True,
        set_first_col_as_index=False,
        quick_db_out_filename=None,
        clean_labels=True,
        to_numeric=None,
    ):
        """Loads tables from the input file
        as a dictionary of python dataframes.

        Recognizes any special table names, such
        as:

            - `run_parameters`, that specifies the output
            path, alongside to some further analysis related
            specifiers.

            - `inputs_from_files`, that specifies a list of
            additional input files of file types: csv, excel, db.
            See examples in the test folders on the master
            branch of the adapter repo for details on the
            structure and labels of the table.

        Parameters:

            create_db: bool
                Write all tables read from input files
                into a run database

            db_flavor: string
                Database type. Currently implemented:
                'sqlite'

            close_db: bool
                True: close the database that got
                created, False: keep the database open

            save_input: bool
                Save initial input file under the output
                folder

            set_first_col_as_index: bool or list of strings
                True: Set index for all tables
                False: do not set the first column as index
                for any tables
                List of strings: List of tables that need
                their first column set as index

            quick_db_out_filename: string, defaults to None
                Output filename without the
                file extension if one does not
                want to use the version tag and outpath
                as provided in a run parameters table when
                saving a database.

                Use with caution as there is no run tag
                or timestamp included. This may be useful when
                quickly converting an excel file with tables
                and named ranges into a database.

            to_numeric: list
                List of string table names where
                values should be converted to
                numeric where possible

            clean_labels: bool
                Process table columns to remove trailing whitespaces

        Returns:

            res : dict
                Keys:

                'tables_as_dict_of_dfs' - all input
                    tables loaded in python as dictionary
                    of dataframes
                'outpath' - output folder path
                'run_tag' - version + analysis start time

                If db got written:

                'db_path' - database fullpath
                'db_conn' - database connection
        """
        dict_of_dfs = self.get_tables(self.input_path)

        # are there any further input files?
        # if that is the case, the file paths and further info
        # should be placed in an `inputs_from_files` table
        qry_flags = dict()

        if self.la["extra_files"] in dict_of_dfs.keys():

            extra_files = dict_of_dfs[self.la["extra_files"]].reset_index()

            for inx in extra_files.index:

                file_path = extra_files.loc[inx, self.la["inpath"]].strip()

                table_names = extra_files.loc[inx, self.la["tbl_nam"]]

                if isinstance(table_names, str):
                    table_names = re.split(",", table_names)
                    table_names = [i.strip() for i in table_names]

                qry_flags[file_path] = extra_files.loc[inx, self.la["query"]]

                if (qry_flags[file_path] is not None) and isinstance(
                    table_names, str
                ):

                    qry_flags[file_path] = re.split(",", qry_flags[file_path])
                    qry_flags[file_path] = [
                        i.strip() for i in qry_flags[file_path]
                    ]

                dict_of_dfs.update(
                    self.get_tables(
                        file_path,
                        table_names=table_names,
                        query_only=qry_flags[file_path],
                        pre_existing_keys=dict_of_dfs.keys(),
                    )
                )

        else:
            qry_flags = None

        # define output path for the analysis run

        if quick_db_out_filename is None:
            # look for `run_parameters` table to extract the outpath
            # and the version
            # `run_parameters` table should occur only in one
            # of the input files, and only once, so if multiple
            # run_parameters{any_text} tables are found, then
            # the first one, when names sorted alphabetically,
            # will be used

            run_params_table = []

            for key in dict_of_dfs.keys():

                if self.la["run_pars"] in key:

                    msg = "Identified run parameters table named {}"
                    log.info(msg.format(key))

                    run_params_table.append(key)

            if len(run_params_table) > 1:
                run_params_table.sort()

                msg = (
                    "Run parameters table named {} will be used to set the "
                    "outpath and the version. "
                    "Additional run parameters table(s) named: {} "
                    "will be ignored. Please make sure to remove any"
                    " unwanted run_parameters tables from the inputs."
                )

                log.warning(
                    msg.format(run_params_table[0], run_params_table[1:])
                )

            if len(run_params_table) != 0:

                outpath_base = os.path.join(
                    os.getcwd(),
                    dict_of_dfs[run_params_table[0]].loc[
                        0, self.la["outpath"]
                    ],
                )

                version = dict_of_dfs[run_params_table[0]].loc[
                    0, self.la["version"]
                ]

                if not isinstance(version, str):
                    # if it was read in as a number, as occurs in the test_input.xlsx on OSX
                    if str(version).endswith(".0"):
                        # Assume that the only case is when version "123" got read in as number "123.0"
                        # period will be removed next
                        version = str(version).rstrip("0")

                # Removing '.', '\', '/' characters from version
                # to avoid any errors during writing output
                version = re.sub("[\\\\./]", "", version)

                self.version = version

            # otherwise declare current folder + "/output" as the output
            # path
            else:
                outpath_base = os.path.join(os.getcwd(), "output")
                version = ""

            run_tag = (
                version + "_" + datetime.now().strftime("%Y_%m_%d-%Hh_%Mm")
            )

            outpath = os.path.join(outpath_base, run_tag)

        else:
            outpath = os.getcwd()
            run_tag = quick_db_out_filename

        if not os.path.exists(outpath):
            os.makedirs(outpath)

        if save_input and isinstance(self.input_path, str):
            # self.input_path
            filename = ntpath.basename(self.input_path)
            filename_extns = re.split("\.", filename)[-1]
            filename_only = re.split("\.", filename)[0]

            versioned_filename = (
                filename_only + "_" + run_tag + "." + filename_extns
            )

            copy(self.input_path, os.path.join(outpath, versioned_filename))

        if create_db == True:

            try:
                db_res = self.create_db(
                    dict_of_dfs,
                    outpath=outpath,
                    run_tag=run_tag,
                    flavor=db_flavor,
                    close=close_db,
                )
            except:
                msg = (
                    "Not able to create a db of tables "
                    "that were read in from {}."
                )

                log.error(msg.format(self.input_path))

        if set_first_col_as_index != False:
            dict_of_dfs = self.first_col_to_index(
                dict_of_dfs, table_names=set_first_col_as_index, drop=True
            )

        # value type conversion for any tables listed
        # as to_numeric and to_float
        if to_numeric is not None:
            if isinstance(to_numeric, list):
                pass
            else:
                msg = (
                    "{} passed for to_numeric kwarg."
                    "Only None or a list of strings are supported."
                )
                log.error(msg.format(to_numeric))

            for key in dict_of_dfs.keys():
                if any([key in tb_nm for tb_nm in to_numeric]):
                    dict_of_dfs[key] = dict_of_dfs[key].apply(
                        pd.to_numeric, errors="ignore", axis=1
                    )

        res = dict()
        res["tables_as_dict_of_dfs"] = dict_of_dfs

        res["outpath"] = outpath
        res["run_tag"] = run_tag

        if create_db == True:
            res.update(db_res)

        if clean_labels == True:

            input_tables_list = res["tables_as_dict_of_dfs"]

            for table in input_tables_list:

                table_columns = input_tables_list[table].columns

                clean_cols = self.process_column_labels(table_columns)
                input_tables_list[table].columns = clean_cols

            msg = "All table column labels were processed to remove undesired whitespaces."
            log.info(msg)

        return res

    def get_tables(
        self,
        file_path,
        table_names=None,
        query_only=None,
        pre_existing_keys=None,
    ):
        """Gets all tables from an input
        file. Creates a dictionary
        of pandas dataframes, with each dataframe
        corresponding to one of the input tables.
        If it is a csv file the contents
        get loaded as a single table with the
        dictionary key being the name of the
        file

        Parameters:

            file_path: str or None
                Input file path
                No data is read in when None,
                creates an empty dictionary

            query_only: str list of 'N' and 'Y', or empty cell (None)
                Default: None - all tables get
                loaded.
                Values in the list: 'N' or 'Y'

            table_names: list of str
                Tables to load. If None
                all tables get loaded (
                unless all need to be
                queried)

            pre_existing_keys: dictionary key index
                Keys is the previously loaded
                dictionary of dataframes

        Returns:

            dict_of_dfs: dict of pd dfs
                Dictionary of pandas dataframes
                containing all the tables
                from the input file, less those
                indicated using the query_only
                flags, if applicable.
        """
        file_path = convert_network_drive_path(file_path)

        if file_path is None:

            dict_of_dfs = dict()

        elif isinstance(file_path, str):

            file_type = self.get_file_type(file_path)

            if isinstance(query_only, str):

                query_only = re.split(",", query_only)
                query_only = [i.strip() for i in query_only]

                if table_names is None:
                    if len(query_only) != 1:
                        msg = (
                            "All tables need to be loaded."
                            "It is unclear which tables need to "
                            "only be queried. Please provide a "
                            "list of table names and query flags "
                            "of the same length."
                        )
                        log.error(msg)

                    elif query_only[0] == "Y":
                        # query all tables
                        table_names_to_load = None
                        table_names_for_conn = table_names

                else:
                    # identify tables that require database connections
                    inx = [i != "Y" for i in query_only]
                    # load only these tables
                    table_names_to_load = np.array(table_names)[inx].tolist()
                    # others should be just connected to
                    not_inx = [not i for i in inx]
                    table_names_for_conn = np.array(table_names)[
                        not_inx
                    ].tolist()

            else:
                table_names_to_load = table_names
                table_names_for_conn = None

            if file_type == "excel":
                # load all tables found in the
                # file as a dict of dataframes

                dict_of_dfs = Excel(file_path, pre_existing_keys).load(
                    data_object_names=table_names_to_load
                )

            elif file_type == "text":
                dict_of_dfs = dict()

                filename_to_tablename = ntpath.basename(file_path)
                filename_to_tablename = re.split("\.", filename_to_tablename)[
                    0
                ]

                Debugger.check_for_duplicates(
                    pre_existing_keys, filename_to_tablename
                )

                # get rid of the version substring
                if self.la["extra_files"] in filename_to_tablename:
                    filename_to_tablename = self.la["extra_files"]

                dict_of_dfs[filename_to_tablename] = pd.read_csv(file_path)

            elif file_type == "database":
                # load all tables found in the
                # file as a dict of dataframes

                dict_of_dfs = Db(file_path, pre_existing_keys).load(
                    table_names=table_names_to_load
                )

            elif file_type == "sqlalchemy":
                # load all tables found in the
                # sqlalchemy database as a dict of dataframes

                dict_of_dfs = Db_sqlalchemy(file_path, pre_existing_keys).load(
                    table_names=table_names_for_conn
                )

        else:
            msg = "Unsupported value ({}) provided as input file path."
            log.error(msg.format(file_path))

        return dict_of_dfs

    def create_db(
        self,
        dict_of_dfs,
        db_conn=False,
        outpath=None,
        run_tag="",
        flavor="sqlite",
        close=True,
    ):
        """Creates a database with all the input
        tables that were read in.

        Parameters:

            dict_of_dfs: dict of dfs
                Contains all input
                tables with table name
                as a dict key and the table
                as a pandas dataframe under that
                key

            db_conn: None or db connection
                None: Writes to the db initiated at
                input data read-in
                Otherwise pass a database connection
                to an existing database

            outpath:
                Output folder path

            run_tag: str
                Default: ""
                Tag to include in the db name

            db_flavor:
                Database file format

        Returns:

            res: dict
                {'db_path' : database path ,
                 'db_conn' : database connection}
        """
        if flavor == "sqlite":
            db_out_type = ".db"

        if outpath is None:
            # create an `output` folder under CWD
            outpath = os.path.join(os.getcwd(), "output")

        if not os.path.exists(outpath):
            os.makedirs(outpath)

        # create an sql database within the output folder and connect
        db_path = os.path.join(outpath, run_tag + db_out_type)
        db_con = sqlite3.connect(db_path)

        # write all input tables in the db
        for table_name in dict_of_dfs.keys():
            try:
                dict_of_dfs[table_name].to_sql(
                    name=table_name,
                    con=db_con,
                    if_exists="replace",
                    index=False,
                )
            except:
                msg = "An error occurred when writting {} table " "to a db {}."
                log.error(msg.format(table_name, db_path))
                raise ValueError

            msg = "Wrote tables in a database: {}."
            log.info(msg.format(db_path))

        if close:
            db_con.close()

        res = {"db_path": db_path, "db_conn": db_con}

        return res

    def write(
        self,
        type="db",
        data_connection=None,
        data_as_dict_of_dfs=None,
        outpath=None,
        run_tag="",
        db_conn=None,
        close_db=True,
    ):
        """Writes all dataframes from a dictionary of dataframes
        out into an existing database.

        Paramters:

            type: str
                String to indicate type of output
                files:

                'db': database
                'csv': csv
                'db&csv': both database and csv

            data_connection: dict
                Return of `load` method

                Keys:

                'tables_as_dict_of_dfs' - all input
                    tables loaded in python as dictionary
                    of dataframes, to be written

                If data_as_dict_of_dfs is not none, these
                will be the dataframes written to output
                instead

                'outpath' - output folder path
                'run_tag' - version + analysis start time


                If db got written:

                'db_path' - database fullpath
                'db_conn' - database connection

                Set to None if passing the data, db
                connection and run tag explicitely through
                other kwargs. This functionality can
                be used when the `load` method was not called.

            data_as_dict_of_dfs: dict of dfs to be written
                Default: None, if data was passed with the
                data_connection

            outpath:
                Default: None means current working directory
                It is ignored if using the data connection
                that is the return of `load` method.
                Otherwise pass output folder path

            run_tag: str
                Default: ""
                It is ignored if using the data connection
                that is the return of `load` method.
                Tag to include in the db/file name

            db_conn: None or db connection
                Default: None if using the data connection that
                is the return of `load` method.
                Otherwise pass a database connection
                to an existing database

            close_db: boolean
                Default: True
                To close the db connection after
                the data is written.

        Return:

            True
        """
        if data_connection is not None:

            if data_as_dict_of_dfs is None:
                data_as_dict_of_dfs = data_connection["tables_as_dict_of_dfs"]

            db_conn = data_connection["db_conn"]

            if outpath is None:
                outpath = data_connection["outpath"]

            run_tag = data_connection["run_tag"]
        else:
            if outpath is None:
                outpath = os.getcwd()

        outpath = convert_network_drive_path(outpath)

        if data_as_dict_of_dfs is None:
            msg = "No data to write passed."
            log.error(msg)
            raise ValueError
        elif not isinstance(data_as_dict_of_dfs, dict):
            msg = "Data needs to be in a " "dictionary of dataframes format."
            log.error(msg)
            raise ValueError

        if "db" in type:

            close = close_db

            if db_conn is None:
                msg = "Missing db connection."
                log.error(msg)
                raise ValueError

            self.create_db(
                dict_of_dfs=data_as_dict_of_dfs,
                db_conn=db_conn,
                outpath=outpath,
                run_tag=run_tag,
                flavor="sqlite",
                close=close_db,
            )

        if "csv" in type:

            if not os.path.exists(outpath):
                os.mkdir(outpath)

            for key in data_as_dict_of_dfs.keys():
                df_to_write = data_as_dict_of_dfs[key]

                df_to_write.to_csv(
                    path_or_buf=os.path.join(
                        outpath,
                        key + "_" + run_tag + ".csv",
                    )
                )

    def first_col_to_index(self, dict_of_dfs, table_names=True, drop=True):
        """Function that sets the first column of dataframe as index.

        Parameters:

            dict_of_dfs: dict of pandas dataframes
                Contains all input
                tables with table name
                as a dict key and the table
                as a pandas dataframe under that
                key

            table_names: list or True
                List containing names of tables that
                need to be modified
                Default: True means all tables are
                modified

            drop: boolean
                Flag indicating whether to drop the column
                being set as index. Default value is true.

        Returns:

            res: dict of pandas dataframes
                Contains the modified dataframes after
                the index has been set

        """
        if table_names == True:
            msg = "Converting first column to index for all inputs."
            log.info(msg.format(table_names))
            table_names = dict_of_dfs.keys()

        elif isinstance(table_names, list):
            msg = "Converting first column to index for {} tables."
            log.info(msg.format(table_names))

        elif table_names == False:
            pass

        else:
            msg = "Unsupported value {} passed to table_names kwarg."
            log.error(msg.format(table_names))
            raise ValueError

        res = dict()
        for x in dict_of_dfs.keys():
            if x in table_names:
                col = dict_of_dfs[x].columns[0]
                res[x] = dict_of_dfs[x].copy().set_index(col, drop=True)
            else:
                res[x] = dict_of_dfs[x].copy()

        return res

    def process_column_labels(self, list_of_labels):
        """
        Converts table columns to string type and removes undesired spaces

        Parameters:

            list_of_labels: list
                list with column labels

        Returns:

            list_of_cleaned_labels: list
                A list with cleaned lables
        """
        list_of_labels = [str(x) for x in list_of_labels]

        list_of_cleaned_labels = [
            re.sub(" +", " ", lbl.strip()) for lbl in list_of_labels
        ]

        return list_of_cleaned_labels



import sys, re


def process_column_labels(list_of_labels):
    """Removes undesired spaces.

    Parameters:

        list_of_labels: list
            list with column labels

    Returns:

        list_of_cleaned_labels: list
            A list with cleaned lables
    """
    list_of_cleaned_labels = [
        re.sub(" +", " ", lbl.strip()) if lbl is str else lbl
        for lbl in list_of_labels
    ]

    return list_of_cleaned_labels


def convert_network_drive_path(
    str_or_path, mapping=[("X:", "/Volumes/my_drive")]
):
    """
    Convert network drive paths from those formatted for one OS into those formatted for another. (works for Windows <-> OSX)
    If a string that doesn't seem to represent a path in the other OS is given, it will be returned unchanged.

    Parameters:
        str_or_path: str
            string holding a filepath.

        mapping: list
            list of 2-tuples where 0th entry of each tuple is the name of a windows network drive location (e.g. "A:") and the 1st entry is OSX network drive location (e.g. "/Volumes/A"). Defaults to [("X:","/Volumes/my_folder")].

    Returns:
        str_or_path: str
            string holding a converted filepath, or original in the case that no mapped network drive was found.

    Raises:
        Exception: When no mapping is given
    """
    if not isinstance(str_or_path, str):
        return str_or_path

    if mapping:
        windows_drive_names = [pair[0].rstrip("\\") for pair in mapping]
        osx_drive_names = [pair[1].rstrip("/") for pair in mapping]
    else:
        raise Exception("No network drive mappings given")

    if sys.platform.startswith("win"):
        for i, name in enumerate(osx_drive_names):
            if str_or_path.startswith(name):
                str_or_path = str_or_path.replace(
                    name, windows_drive_names[i]
                ).replace("/", "\\")
                break

    elif sys.platform.startswith("darwin"):
        for i, name in enumerate(windows_drive_names):
            if str_or_path.startswith(name):
                str_or_path = str_or_path.replace("\\", "/").replace(
                    name, osx_drive_names[i]
                )
                break

    return str_or_path



#!/usr/bin/env python

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import codecs
import os.path

# to access version
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

long_description = "The `Adapter Python IO` software provides a convenient data table loader from various formats such as `xlsx`, `csv`, `db (sqlite database)`, and `sqlalchemy`. Its main feature is the ability to convert data tables identified in one main and optionally one or more additional input files into `database tables` and `Pandas DataFrames` for downstream usage in any compatible software. `Adapter` builds upon the existing Python packages that allow for the communication between `Python` and `MS Excel`, as well as `databases` and `csv` files. It provides inbuilt capabilities to specify the output location path, as well as a version identifier for a research code run.  In addition to the loading capability, an instance of the `Adapter` `IO` object has the write capability. If invoked, all loaded tables are written as either a single `database` or a set of `csv` files, or both. The purpose of this software is to support the development of research and analytical software through allowing for a simple multi-format IO with versioning and output path specification in the input data itself. The package is supported on `Windows` and `macOS`, as well as for `Linux` for the utilization without any `xlsx` inputs."

setup(
    name="adapterio",
    version=get_version("adapter/__init__.py"),
    description="I/O module",
    long_description=long_description,
    # The project's main homepage.
    url="https://github.com/LBNL-ETA/Adapter",
    # Author details
    author="Milica Grahovac, Youness Bennani, Thomas Burke, Katie Coughlin, Mohan Ganeshalingam, Akhil Mathur, Evan Neill, and Akshay Sharma",
    # Choose your license
    license="BSD-3-Clause-LBNL",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Programming Language :: Python :: 3.8",
    ],
    keywords="data, tables, IO for research computation, sql, excel, csv, dataframe, connection",
    packages=find_packages(exclude=["*.tests", "*.tests"]),
    install_requires=[
        "pandas>=1.0.4",
        "xlwings>=0.19.4",
        "psycopg2-binary>=2.8.6",
    ],
)
